/*
 * This program is to demonstrate application of Runge-Kutta methods
 * for solution of Cauchy problem for ordinary differential equation
 * in one dimension with fixed step size.
 *
 * (C) vladimir.smirnov.penza@gmail.com (ftp://95.83.126.89), 2010
 * Redistribution in source or binary form, modified or not, permitted
 * under the terms of BSD license.
 *
 * To compile & run:
 *  - Windows: install MinGW(/Cygwin?), than say make.
 *  - GNU environment (Linux, notably): just say make.
 *  - generic UN*X: try to say gmake (you'll need GCC and GNU make).
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/*
 * Maximum number of points to output
 */
#define OUTPTS  20

/* -----------------------------------------------------------------------
 * Problem-independent part.
 * Step functions for 4 methods from Runge-Kutta group together with
 * simply driver routine (fixed-size grid).
 */

/*
 * 1st orger: Euler's method.
 */
static void
rk1(double *x,double *y, /* point to update */
	double (*dydx)(double,double), /* right side of the equation */
	double dx /* step of integration */ )
{
	*y += dydx(*x, *y) * dx;
	*x += dx;
}

/*
 * 2nd orger: midpoint method.
 */
static void
rk2(double *x,double *y, /* point to update */
	double (*dydx)(double,double), /* right side of the equation */
	double dx /* step of integration */ )
{
	double k1, halfdx = 0.5 * dx;
	/*
	 * Method of second order requires two computations of right side.
	 */
	k1 = *y + dydx(*x, *y) * halfdx;
	*y += dydx(*x + halfdx, k1) * dx;
	*x += dx;
}

/*
 * Another 2nd orger: predictor-corrector scheme.
 * Not very usefull in most cases.
 */
static void
rk2pc(double *x,double *y, /* point to update */
	double (*dydx)(double,double), /* right side of the equation */
	double dx /* step of integration */ )
{
	double k1 = dydx(*x, *y);
	*x += dx;
	*y += 0.5 * (k1 + dydx(*x, k1 * dx)) * dx;
}

/*
 * 4th orger: classical RK4.
 */
static void
rk4(double *x,double *y, /* point to update */
	double (*dydx)(double,double), /* right side of the equation */
	double dx /* step of integration */ )
{
	static const double sixth = 1.0 / 6.0;
	double k1, k2, k3, k4;
	double halfdx, halfx;
	/*
	 * Method of fourth order requires four computations of right side.
	 */
	halfdx = 0.5 * dx;
	halfx = *x + halfdx;

	k1 = dydx(*x, *y);
	k2 = dydx(halfx, *y + k1 * halfdx);
	k3 = dydx(halfx, *y + k2 * halfdx);
	k4 = dydx(*x + dx, *y + k3 * dx);
	*y += sixth * (k1 + 2.0 * (k2 + k3) + k4) * dx;
	*x += dx;
}

typedef struct stepfuncdesc_s
{
	char *id;
	char *desc;
	void (*stepfunc)(double *x,double *y, /* point to update */
		double (*dydx)(double x,double y), /* right side of the equation */
		double dx); /* step of integration */
} stepfuncdesc_t, *stepfuncdesc_p;

static stepfuncdesc_t known_methods[] =
{
	{"euler", "Runge-Kutta of 1st order (Euler's method)", rk1},
	{"rk2", "Runge-Kutta of 2dn order (midpoint method)", rk2},
	{"rk2pc", "Runge-Kutta of 2nd order (predictor - corrector)", rk2pc},
	{"rk4", "Runge-Kutta of 4th order (classic RK4)", rk4},
	{NULL, NULL, NULL}
};

static void
list_methods(void)
{
	stepfuncdesc_p method = known_methods;
	for(;method->id;method++)
		printf("Method: \"%s\": %s.\n", method->id, method->desc);
}

static stepfuncdesc_p
find_method(char *id)
{
	stepfuncdesc_p method = known_methods;
	for(;method->id;method++)
		if(!strcmp(id, method->id))
			return method;
	return NULL;
}

/*
 * Driver routine: common for every methods.
 */
static double
Cauchy(
	/* initial conditions */
	double x0, double y0,
	/* selected solver routine for updating current point */
	stepfuncdesc_p method,
	/* right side of the equation to solve */
	double (*dydx)(double,double),double dx, double xn)
{
	unsigned i, skip, everypoint;
	void (*stepfunc)(double*,double*,double (*dydx)(double,double),double);
	double x = x0, y = y0;
	skip = (unsigned)(floor((xn - x0) / dx)) / OUTPTS;
	everypoint = skip < 2;
	stepfunc = method->stepfunc;
	for(i = 0;;i++)
	{
		if(everypoint || !(i % skip))
			printf("%.5g;%.10g\n", x, y);
		(*stepfunc)(&x, &y, dydx, dx);
		if(x + dx <= xn)
			continue;
		dx = xn - x;
		(*stepfunc)(&x, &y, dydx, dx);
		printf("%.5g;%.10g\n", x, y);
		break;
	}
	return y;
}

/* -----------------------------------------------------------------------
 * Problem-dependent part.
 */

/*
 * Let's solve Cauchy problem for the equation $\frac{dy}{dx}=-xy$.
 * This equation is separable, the analytic solution is
 * $$y=C e^{- \frac{x^2}{2}}$$.
 */
__inline double rightside(double  x,double y) { return -x * y; }
/*
 * Initial conditions
 */
#define X0  (double)(-2.0)
#define Y0  (double)(1 * exp(- X0 * X0 / 2))    /* (app. 0.135) */
#define Xn  (double)(2.0)

int
main(int argc, char *argv[])
{
	double step;
	char *endptr = NULL;
	stepfuncdesc_p method;
	if(argc != 3)
	{
		printf("Usage: %s <method> <step of integration>\n"
			"Known methods:\n", argv[0]);
		list_methods();
		return -1;
	}
	method = find_method(argv[1]);
	if(!method)
	{
		printf("Uknkown method \"%s\".\nKnown methods:\n", argv[1]);
		list_methods();
		return -1;
	}
	step = strtod(argv[2], &endptr);
	if(!endptr || *endptr)
	{
		printf("Third parameter must be valid floating-point value.\n");
		return -1;
	}
	if(step < 0)
	{
		printf("Invlid step of integration \"%g\".\n", step);
		return -1;
	}
	/*
	 * Compute integral curve up to the point with x = 2.
	 * The third parameter may be rk[1|2|4].
	 */
	Cauchy(X0, Y0, method, rightside, step, Xn);
	return 0;
}
